//
//  ProfileTableViewController.swift
//  Tusker
//
//  Created by Shadowfacts on 8/27/18.
//  Copyright © 2018 Shadowfacts. All rights reserved.
//

import UIKit
import Pachyderm
import SafariServices

class ProfileTableViewController: EnhancedTableViewController {
    
    weak var mastodonController: MastodonController!
    
    var accountID: String! {
        didSet {
            if shouldLoadOnAccountIDSet {
                DispatchQueue.main.async {
                    self.updateAccountUI()
                }
            }
        }
    }
    
    var pinnedStatuses: [(id: String, state: StatusState)] = [] {
        didSet {
            DispatchQueue.main.async {
                self.tableView.reloadData()
            }
        }
    }
    var timelineSegments: [[(id: String, state: StatusState)]] = [] {
        didSet {
            DispatchQueue.main.async {
                self.tableView.reloadData()
            }
        }
    }
    
    var older: RequestRange?
    var newer: RequestRange?
    
    var shouldLoadOnAccountIDSet = false
    var loadingVC: LoadingViewController? = nil
    
    init(accountID: String?, mastodonController: MastodonController) {
        self.mastodonController = mastodonController
        
        self.accountID = accountID
        
        super.init(style: .plain)
        
        self.refreshControl = UIRefreshControl()
        refreshControl!.addTarget(self, action: #selector(refreshStatuses(_:)), for: .valueChanged)
        navigationItem.rightBarButtonItem = UIBarButtonItem(barButtonSystemItem: .compose, target: self, action: #selector(composePressed(_:)))
    }
    
    required init?(coder aDecoder: NSCoder) {
        fatalError("init(coder:) has not been implemeneted")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()

        tableView.rowHeight = UITableView.automaticDimension
        tableView.estimatedRowHeight = 140
        
        tableView.register(UINib(nibName: "TimelineStatusTableViewCell", bundle: nil), forCellReuseIdentifier: "statusCell")
        tableView.register(UINib(nibName: "ProfileHeaderTableViewCell", bundle: nil), forCellReuseIdentifier: "headerCell")
        
        tableView.prefetchDataSource = self
        
        if let accountID = accountID {
            if mastodonController.cache.account(for: accountID) != nil {
                updateAccountUI()
            } else {
                loadingVC = LoadingViewController()
                embedChild(loadingVC!)
                mastodonController.cache.account(for: accountID) { (account) in
                    guard account != nil else {
                        let alert = UIAlertController(title: "Something Went Wrong", message: "Couldn't load the selected account", preferredStyle: .alert)
                        alert.addAction(UIAlertAction(title: "OK", style: .default, handler: { (_) in
                            self.navigationController!.popViewController(animated: true)
                        }))
                        DispatchQueue.main.async {
                            self.present(alert, animated: true)
                        }
                        return
                    }
                    DispatchQueue.main.async {
                        self.updateAccountUI()
                        self.tableView.reloadData()
                    }
                }
            }
        } else {
            loadingVC = LoadingViewController()
            embedChild(loadingVC!)
            shouldLoadOnAccountIDSet = true
        }
        
        NotificationCenter.default.addObserver(self, selector: #selector(updateUIForPreferences), name: .preferencesChanged, object: nil)
    }

    func updateAccountUI() {
        loadingVC?.removeViewAndController()
        
        updateUIForPreferences()
        
        getStatuses(onlyPinned: true) { (response) in
            guard case let .success(statuses, _) = response else { fatalError() }
            
            self.mastodonController.cache.addAll(statuses: statuses)
            self.pinnedStatuses = statuses.map { ($0.id, .unknown) }
        }
        
        getStatuses() { response in
            guard case let .success(statuses, pagination) = response else { fatalError() }
            
            self.mastodonController.cache.addAll(statuses: statuses)
            self.timelineSegments.append(statuses.map { ($0.id, .unknown) })
            
            self.older = pagination?.older
            self.newer = pagination?.newer
        }
    }
    
    @objc func updateUIForPreferences() {
        guard let accountID = accountID, let account = mastodonController.cache.account(for: accountID) else { return }
        navigationItem.title = account.realDisplayName
    }
    
    func getStatuses(for range: RequestRange = .default, onlyPinned: Bool = false, completion: @escaping Client.Callback<[Status]>) {
        let request = Account.getStatuses(accountID, range: range, onlyMedia: false, pinned: onlyPinned, excludeReplies: !Preferences.shared.showRepliesInProfiles)
        mastodonController.run(request, completion: completion)
    }
    
    func sendMessageMentioning() {
        guard let account = mastodonController.cache.account(for: accountID) else { fatalError("Missing cached account \(accountID!)") }
        let vc = UINavigationController(rootViewController: ComposeViewController(mentioningAcct: account.acct, mastodonController: mastodonController))
        present(vc, animated: true)
    }
    
    // MARK: - Table view data source

    override func numberOfSections(in tableView: UITableView) -> Int {
        // 1 section for header, 1 section for pinned, rest for timeline
        return 2 + timelineSegments.count
    }

    override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if section == 0 {
            return accountID == nil || mastodonController.cache.account(for: accountID) == nil ? 0 : 1
        } else if section == 1 {
            return pinnedStatuses.count
        } else {
            return timelineSegments[section - 2].count
        }
    }

    override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        switch indexPath.section {
        case 0:
            guard let cell = tableView.dequeueReusableCell(withIdentifier: "headerCell", for: indexPath) as? ProfileHeaderTableViewCell else { fatalError() }
            cell.selectionStyle = .none
            cell.delegate = self
            cell.updateUI(for: accountID)
            return cell
        case 1:
            guard let cell = tableView.dequeueReusableCell(withIdentifier: "statusCell", for: indexPath) as? TimelineStatusTableViewCell else { fatalError() }
            let (id, state) = pinnedStatuses[indexPath.row]
            cell.showPinned = true
            cell.delegate = self
            cell.updateUI(statusID: id, state: state)
            return cell
        default:
            guard let cell = tableView.dequeueReusableCell(withIdentifier: "statusCell", for: indexPath) as? TimelineStatusTableViewCell else { fatalError() }
            let (id, state) = timelineSegments[indexPath.section - 2][indexPath.row]
            cell.delegate = self
            cell.updateUI(statusID: id, state: state)
            return cell
        }
    }
    
    // MARK: - Table view delegate

    override func tableView(_ tableView: UITableView, willDisplay cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        if timelineSegments.count > 0 && indexPath.section - 1 == timelineSegments.count && indexPath.row == timelineSegments[indexPath.section - 2].count - 1 {
            guard let older = older else { return }
            
            getStatuses(for: older) { response in
                guard case let .success(newStatuses, pagination) = response else { fatalError() }
                
                self.mastodonController.cache.addAll(statuses: newStatuses)
                self.timelineSegments[indexPath.section - 2].append(contentsOf: newStatuses.map { ($0.id, .unknown) })

                self.older = pagination?.older
            }
        }
    }

    override func tableView(_ tableView: UITableView, canEditRowAt indexPath: IndexPath) -> Bool {
        return true
    }
    
    override func tableView(_ tableView: UITableView, leadingSwipeActionsConfigurationForRowAt indexPath: IndexPath) -> UISwipeActionsConfiguration? {
        return (tableView.cellForRow(at: indexPath) as? TableViewSwipeActionProvider)?.leadingSwipeActionsConfiguration()
    }
    
    override func tableView(_ tableView: UITableView, trailingSwipeActionsConfigurationForRowAt indexPath: IndexPath) -> UISwipeActionsConfiguration? {
        return (tableView.cellForRow(at: indexPath) as? TableViewSwipeActionProvider)?.trailingSwipeActionsConfiguration()
    }
    
    @objc func refreshStatuses(_ sender: Any) {
        guard let newer = newer else { return }
        
        getStatuses(for: newer) { response in
            guard case let .success(newStatuses, pagination) = response else { fatalError() }
            
            self.mastodonController.cache.addAll(statuses: newStatuses)
            self.timelineSegments[0].insert(contentsOf: newStatuses.map { ($0.id, .unknown) }, at: 0)

            if let newer = pagination?.newer {
                self.newer = newer
            }

            DispatchQueue.main.async {
                self.refreshControl?.endRefreshing()
            }
        }
        
        getStatuses(onlyPinned: true) { (response) in
            guard case let .success(newPinnedStatuses, _) = response else { fatalError() }
            self.mastodonController.cache.addAll(statuses: newPinnedStatuses)
            
            let oldPinnedStatuses = self.pinnedStatuses
            var pinnedStatuses = [(id: String, state: StatusState)]()
            for status in newPinnedStatuses {
                let state: StatusState
                if let (_, oldState) = oldPinnedStatuses.first(where: { $0.id == status.id }) {
                    state = oldState
                } else {
                    state = .unknown
                }
                pinnedStatuses.append((status.id, state))
            }
            self.pinnedStatuses = pinnedStatuses
        }
    }
    
    @objc func composePressed(_ sender: Any) {
        sendMessageMentioning()
    }
    
}

extension ProfileTableViewController: StatusTableViewCellDelegate {
    var apiController: MastodonController { mastodonController }
    
    func statusCellCollapsedStateChanged(_ cell: BaseStatusTableViewCell) {
        // causes the table view to recalculate the cell heights
        tableView.beginUpdates()
        tableView.endUpdates()
    }
}

extension ProfileTableViewController: ProfileHeaderTableViewCellDelegate {
    func showMoreOptions(cell: ProfileHeaderTableViewCell) {
        let account = mastodonController.cache.account(for: accountID)!
        
        mastodonController.cache.relationship(for: account.id) { [weak self] (relationship) in
            guard let self = self else { return }
            
            var customActivities: [UIActivity] = [OpenInSafariActivity()]
            if let relationship = relationship {
                let toggleFollowActivity = relationship.following ? UnfollowAccountActivity() : FollowAccountActivity()
                customActivities.insert(toggleFollowActivity, at: 0)
            }
            
            DispatchQueue.main.async {
                let activityController = UIActivityViewController(activityItems: [account.url, account], applicationActivities: customActivities)
                activityController.completionWithItemsHandler = OpenInSafariActivity.completionHandler(viewController: self, url: account.url)
                activityController.popoverPresentationController?.sourceView = cell.moreButtonVisualEffectView
                self.present(activityController, animated: true)
            }
        }
    }
}

extension ProfileTableViewController: UITableViewDataSourcePrefetching {
    func tableView(_ tableView: UITableView, prefetchRowsAt indexPaths: [IndexPath]) {
        for indexPath in indexPaths where indexPath.section > 1 {
            let statusID = timelineSegments[indexPath.section - 2][indexPath.row].id
            guard let status = mastodonController.cache.status(for: statusID) else { continue }
            _ = ImageCache.avatars.get(status.account.avatar, completion: nil)
            for attachment in status.attachments {
                _ = ImageCache.attachments.get(attachment.url, completion: nil)
            }
        }
    }
    
    func tableView(_ tableView: UITableView, cancelPrefetchingForRowsAt indexPaths: [IndexPath]) {
        for indexPath in indexPaths where indexPath.section > 1 {
            let statusID = timelineSegments[indexPath.section - 2][indexPath.row].id
            guard let status = mastodonController.cache.status(for: statusID) else { continue }
            ImageCache.avatars.cancelWithoutCallback(status.account.avatar)
            for attachment in status.attachments {
                ImageCache.attachments.cancelWithoutCallback(attachment.url)
            }
        }
    }
}
